#!/usr/bin/env python3
# -*-coding:utf-8-*-

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional, Type

from loguru import logger

from .llm import LLM, GenericResponseType


@dataclass
class ReAct(ABC):
    thought: Optional[str] = None
    action: Optional[str] = None
    params: Optional[Dict[str, str]] = None
    current: bool = True

    def __post_init__(self):
        self.obs = self.run()

    @abstractmethod
    def run(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def parse(cls, response) -> "ReAct":
        pass

    @property
    @abstractmethod
    def done(self) -> bool:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @property
    def info(self) -> str:
        if self.done:
            return f"[Thought]: {self.thought}"
        else:
            return f"[Thought]: {self.thought}\n[Action]: {self.action}\n[Params]: {self.params}\n[Obs]: {self.obs}"


@dataclass
class ReActChain(LLM[GenericResponseType]):
    reactType: Optional[Type[ReAct]] = None

    def __post_init__(self):
        self._acts = []
        super().__post_init__()
        assert self.reactType is not None, "cls must be set"

    @property
    def input_keys(self) -> List[str]:
        return ["prompt"]

    def _process(self, act: ReAct) -> str:
        logger.success(act.info)
        while not act.done:
            if self._acts:
                self._acts[-1].current = False
            self._acts.append(act)
            inputs = {"prompt": "\n".join(str(a) for a in self._acts)}
            response = self._call(inputs)
            act = self.reactType.parse(response)  # type: ignore
            logger.success(act.info)
        return act.info

    def __call__(self, act: Optional[ReAct] = None) -> str:
        if act is None:
            inputs = {"prompt": "Start!"}
            inputs = self._prep_inputs(inputs)
            response = self._call(inputs)
            act = self.reactType.parse(response)  # type: ignore
        return self._process(act)



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Union

from loguru import logger

from .llm import LLM, FunctionCall, GenericResponseType


@dataclass
class ReduceChain(LLM[GenericResponseType]):
    prompt_template: str = "{history}\n{input}"
    reduce_keys: List[str] = field(default_factory=lambda: ["input"])
    format: Callable[[Union[str, GenericResponseType]], str] = field(default_factory=lambda: lambda x: str(x))

    def __post_init__(self):
        super().__post_init__()
        if self.tools:
            self._format = self.format

            def format_for_tools(response: FunctionCall) -> str:
                for tool in self.tools:  # type: ignore
                    if tool.__name__ == response.name:
                        result = tool(**response.args)
                        logger.debug(f"{tool.__name__}({response.args}) -> {result}")
                        return self._format(result)
                raise ValueError(f"Cannot find tool {response.name}")

            self.format = format_for_tools  # type: ignore

    def _validate_inputs(self, inputs: Dict[str, Any]) -> None:
        self.num = len(inputs[self.reduce_keys[0]])
        for reduce_key in self.reduce_keys:
            self._validate_reduce_key(reduce_key, inputs)
        assert "history" in self.input_keys, "ReduceChain expects response to be in input_keys"
        if "history" not in inputs:
            inputs["history"] = ""
        super()._validate_inputs(inputs)

    def _validate_reduce_key(self, reduce_key: str, inputs: Dict[str, Any]) -> None:
        assert reduce_key in self.input_keys, f"ReduceChain expects {reduce_key} to be in input_keys"
        assert isinstance(inputs[reduce_key], List), f"ReduceChain expects {reduce_key} to be a list of strings"
        assert (
            len(inputs[reduce_key]) == self.num
        ), f"ReduceChain expects {reduce_key} to be a list of strings with the same length"

    def _call(self, inputs: Dict[str, str]) -> Union[str, GenericResponseType]:
        response = ""
        history = inputs["history"]
        for i in range(self.num):
            response = self._process(i, inputs, history)
            history = self.format(response)
            logger.debug(f"ReduceChain: {i} finished")
            logger.debug(f"ReduceChain: {response}")
        return response

    def _process(self, index: int, inputs: Dict[str, str], history: str) -> Union[str, GenericResponseType]:
        new_input = inputs.copy()
        if index > 0:
            new_input.pop("history")
            new_input["history"] = history
        new_input.update({reduce_key: inputs[reduce_key][index] for reduce_key in self.reduce_keys})
        return super()._call(new_input)



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from dataclasses import dataclass, field
from typing import Callable, Dict, List, Union

from loguru import logger

from uglychain.llm.tools import ActionResopnse

from .llm import LLM, FunctionCall, GenericResponseType


def finish(answer: str)-> str:
    """returns the answer and finishes the task.
    Args:
        answer (str): The response to return.
    """
    return answer

def call(tools: List[Callable], response: FunctionCall):
    for tool in tools:
        if tool.__name__ == response.name:
            return tool(**response.args)
    raise ValueError(f"Can't find tool {response.name}")

@dataclass
class Action:
    thought: str
    action: str
    params: Dict
    obs: str
    current: bool = True

    @property
    def done(self) -> bool:
        if self.action == "finish":
            return True
        else:
            return False

    def __str__(self) -> str:
        return self.info

    @property
    def info(self) -> str:
            return f"Thought: {self.thought}\nAction: {self.action}\nAction Input: {self.params}\nObservation: {self.obs}"

@dataclass
class ReActChain(LLM[GenericResponseType]):
    llmchain: LLM = field(init=False)

    def __post_init__(self):
        self._acts = []
        #super().__post_init__()
        if self.system_prompt:
            self.llm.set_system_prompt(self.system_prompt)
        self.prompt = self.prompt_template
        assert self.tools is not None, "tools must be set"
        self.tools.insert(0, finish)
        self.llmchain = LLM("""Question: {input}\n{react_history}""",self.model, tools=self.tools, response_model=ActionResopnse)
        self.llmchain.llm.use_native_tools = False

    def _call(self, inputs: Dict[str, str]) -> Union[str, GenericResponseType]:
        input = self.prompt.format(**inputs)
        response = self.llmchain(input = input, react_history = "")
        thought = response.thought
        action = response.action.name
        params = response.action.args
        obs = call(self.tools, response.action)
        act = Action(thought, action, params, obs)
        logger.success(act.info)
        while not act.done:
            if self._acts:
                self._acts[-1].current = False
            self._acts.append(act)
            react_history = "\n".join(str(a) for a in self._acts) + "\n\nWhat do you think to do next? Don't repeat yourself."
            logger.debug(f"{input}\n{react_history}")
            response = self.llmchain(input = input, react_history = react_history)
            thought = response.thought
            action = response.action.name
            params = response.action.args
            obs = call(self.tools, response.action)
            act = Action(thought, action, params, obs)
            logger.success(act.info)
        response = obs
        if self.response_model:
            llm = LLM(model=self.model, response_model=self.response_model)
            return llm(response)
        else:
            return response



#!/usr/bin/env python3
# -*-coding:utf-8-*-


class _SingletonWrapper:
    """A singleton wrapper class.

    Instances of this class are created for each decorated class.
    """

    def __init__(self, cls):
        self.__wrapped__ = cls
        self._instance = None

    def __call__(self, *args, **kwargs):
        """Returns a single instance of the decorated class"""
        if self._instance is None:
            self._instance = self.__wrapped__(*args, **kwargs)
        return self._instance


def singleton(cls):
    """A singleton decorator.

    Returns a wrapper object. A call on that object returns a single instance
    object of the decorated class. Use the __wrapped__ attribute to access the
    decorated class directly in unit tests.
    """
    return _SingletonWrapper(cls)



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from .base import BaseLanguageModel
from .instructor import Instructor, ParseError
from .model import Model
from .tools import FunctionCall

__all__ = [
    "BaseLanguageModel",
    "Model",
    "Instructor",
    "ParseError",
    "FunctionCall",
]



#!/usr/bin/env python3
# -*-coding:utf-8-*-

import re
from dataclasses import dataclass


@dataclass
class Prompt:
    """A class representing a prompt.

    Attributes:
        template: The template string for the prompt.
    """

    template: str = ""

    def format(self, **kwargs):
        """Format the prompt with the given kwargs.

        Args:
            **kwargs: Keyword arguments to replace placeholders in the template.

        Returns:
            The formatted prompt string.
        """

        def replace(match):
            key = match.group(1)
            value = kwargs.get(key, match.group(0))
            if isinstance(value, list):
                return "".join(str(x) for x in value)
            else:
                return str(value)

        pattern = r"(?<!{){([^{}\n]+)}(?!})"
        result = re.sub(pattern, replace, self.template)
        return result

    @property
    def input_variables(self):
        """Get a list of input variables in the template.

        Returns:
            A list of input variables.
        """
        return re.findall(r"(?<!{){([^{}\n]+)}(?!})", self.template)



from enum import Enum

from uglychain.utils import config


class Model(Enum):
    GPT3_TURBO = "gpt-3.5-turbo"
    GPT3_TURBO_16K = "gpt-3.5-turbo-16k"
    GPT4 = "gpt-4"
    GPT4_TURBO = "gpt-4-turbo"
    COPILOT3_TURBO = "copilot-3.5"
    COPILOT4 = "copilot-4"
    YI = "yi"
    YI_32K = "yi-32k"
    BAICHUAN_TURBO = "baichuan-turbo"
    BAICHUAN_TURBO_192K = "baichuan-turbo-192k"
    BAICHUAN_PRO = "baichuan-pro"
    CUSTOM = "custom"
    GEMINI = "gemini"
    QWEN = "qwen"
    QWEN_28K = "qwen-28k"
    QWEN_TURBO = "qwen-turbo"
    QWEN_PLUS = "qwen-plus"
    GLM4 = "glm-4"
    GLM3 = "glm-3"
    DEFAULT = config.llm_provider



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Type

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config, retry_decorator


@dataclass
class ChatGLM(BaseLanguageModel):
    model: str
    use_max_tokens: bool = False
    MAX_TOKENS: int = 128000

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools)
        response = self.completion_with_backoff(**kwargs)
        logger.trace(f"kwargs:{kwargs}\nresponse:{response}")
        return response.choices[0].message.content.strip()

    @property
    def default_params(self) -> Dict[str, Any]:
        kwargs = {
            "model": self.model,
            "do_sample": True,
            "temperature": self.temperature,
        }
        if self.use_max_tokens:
            kwargs["max_tokens"] = self.max_tokens
        return kwargs

    def _create_client(self):
        try:
            from zhipuai import ZhipuAI
        except ImportError as err:
            raise ImportError("You need to install `pip install dashscope` to use this provider.") from err
        return ZhipuAI(api_key=config.zhipuai_api_key)

    @retry_decorator()
    def completion_with_backoff(self, **kwargs):
        return self.client.chat.completions.create(**kwargs)

    @property
    def max_tokens(self):
        tokens = self._num_tokens(messages=self.messages, model=self.model) + 1000  # add 1000 tokens for answers
        if not self.MAX_TOKENS > tokens:
            raise Exception(
                f"Prompt is too long. This model's maximum context length is {self.MAX_TOKENS} tokens. your messages required {tokens} tokens"
            )
        max_tokens = self.MAX_TOKENS - tokens + 1000
        return 2000 if max_tokens > 2000 else max_tokens



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Type

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config, retry_decorator


@dataclass
class Gemini(BaseLanguageModel):
    model: str
    use_max_tokens: bool = False

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools)
        response = self.completion_with_backoff(**kwargs)
        logger.trace(f"kwargs:{kwargs}\nresponse:{response}")
        return response.choices[0].message.content.strip()

    def get_kwargs(
        self,
        prompt: str,
        response_model: Optional[Type],
        tools: Optional[List[Callable]],
    ) -> Dict[str, Any]:
        kwargs = super().get_kwargs(prompt, response_model, tools)
        contents = kwargs["messages"]
        kwargs.pop("messages")
        kwargs["contents"] = contents
        return kwargs

    def _generate_messages(self, prompt: str) -> List[Dict[str, str]]:
        """Generate the list of messages for the conversation.

        Args:
            prompt (str): The user prompt.

        Returns:
            List[Dict[str, str]]: The list of messages.

        """
        if not self.is_continuous:
            self.messages = []
            if hasattr(self, "system_prompt") and self.system_prompt:
                self.messages.append({"role": "system", "parts": [self.system_prompt]})
        if prompt:
            self.messages.append({"role": "user", "parts": [prompt]})
        return self.messages

    @property
    def default_params(self) -> Dict[str, Any]:
        config = self.client.types.GenerationConfig(max_output_tokens=self.max_tokens, temperature=self.temperature)
        kwargs = {
            "generation_config": config,
        }
        return kwargs

    def _create_client(self):
        try:
            import google.generativeai as genai

            genai.configure(api_key=config.gemini_api_key, transport="rest")
        except ImportError as e:
            raise ImportError("You need to install `pip install google-generativeai` to use this provider.") from e
        return genai

    @retry_decorator()
    def completion_with_backoff(self, **kwargs):
        model = self.client.GenerativeModel(self.model)
        return model.generate_content(**kwargs)

    @property
    def max_tokens(self):
        return 2000



import inspect
from enum import Enum
from typing import Callable, List, Union, cast

from docstring_parser import parse
from pydantic import BaseModel, Field


class FunctionCall(BaseModel):
    name: str = Field(..., description="tool name")
    args: dict = Field(..., description="tool arguments")

class ActionResopnse(BaseModel):
    thought: str = Field(..., description="Think step by step and explan why you need to use a tool")
    action: FunctionCall = Field(..., description="The action to take")

FUNCTION_CALL_FORMAT = """
-----
Respond with tool name and tool arguments to achieve the instruction:

{tool_schema}
"""


def to_json_schema_type(type_name: str) -> str:
    type_map = {
        "str": "string",
        "int": "integer",
        "float": "number",
        "bool": "boolean",
        "None": "null",
        "Any": "any",
        "Dict": "object",
        "List": "array",
        "Optional": "any",
    }
    return type_map.get(type_name, "any")


def parse_annotation(annotation):
    if getattr(annotation, "__origin__", None) == Union:
        types = [t.__name__ if t.__name__ != "NoneType" else "None" for t in annotation.__args__]
        return to_json_schema_type(types[0])
    elif issubclass(annotation, Enum):  # If the annotation is an Enum type
        return "enum", [item.name for item in annotation]  # Return 'enum' and a list of the names of the enum members
    elif getattr(annotation, "__origin__", None) is not None:
        if annotation._name is not None:
            return f"{to_json_schema_type(annotation._name)}[{','.join([to_json_schema_type(i.__name__) for i in annotation.__args__])}]"
        else:
            return f"{to_json_schema_type(annotation.__origin__.__name__)}[{','.join([to_json_schema_type(i.__name__) for i in annotation.__args__])}]"
    else:
        return to_json_schema_type(annotation.__name__)


def get_pydantic_schema(pydantic_obj: BaseModel, visited_models=None) -> dict:
    if visited_models is None:
        visited_models = set()

    if pydantic_obj in visited_models:
        raise ValueError(f"Circular reference detected: {pydantic_obj.__name__}")

    visited_models.add(pydantic_obj)

    schema = pydantic_obj.model_json_schema()
    definitions = schema.pop("definitions", {})

    def resolve_schema(schema):
        if "$ref" in schema:
            ref_path = schema["$ref"]
            definition_key = ref_path.split("/")[-1]
            return resolve_schema(definitions[definition_key])
        elif "items" in schema:
            schema["items"] = resolve_schema(schema["items"])
        return schema

    schema = resolve_schema(schema)
    for name, property in schema["properties"].items():
        schema["properties"][name] = resolve_schema(property)

    visited_models.remove(pydantic_obj)

    return schema


def function_schema(func: Callable):
    signature = inspect.signature(func)
    docstring = inspect.getdoc(func)
    if docstring:
        docstring_parsed = parse(docstring)
    else:
        docstring_parsed = parse("")

    parameters = {}
    required = []

    for name, param in signature.parameters.items():
        json_type = parse_annotation(param.annotation)
        param_info = None

        if isinstance(param.annotation, type) and issubclass(param.annotation, BaseModel):
            # If the parameter is a Pydantic object
            # val_func = validate_arguments(func)
            param_info = get_pydantic_schema(cast(BaseModel, param.annotation))
            param_info["description"] = param.annotation.__doc__
            # Add Pydantic object parameter to the required list
            required.append(name)
        elif isinstance(json_type, tuple) and json_type[0] == "enum":  # If the type is an Enum
            param_info = {
                "type": "string",
                "enum": json_type[1],  # Add an 'enum' field with the names of the enum members
                "description": "",
            }
        else:
            param_info = {"type": json_type, "description": ""}

        if json_type != "any" and name != "self" and param.default == inspect.Parameter.empty:
            required.append(name)

        for doc_param in docstring_parsed.params:
            if doc_param.arg_name == name:
                param_info["description"] = doc_param.description if doc_param.description else ""

        parameters[name] = param_info

    function_info = {
        "name": func.__name__,
        "description": docstring_parsed.short_description,
        "parameters": {
            "type": "object",
            "properties": parameters,
            "required": required,
        },
    }

    return function_info


def tools_schema(tools: List[Callable]):
    tools_schema = []

    for tool in tools:
        tool_schema = {"type": "function", "function": function_schema(tool)}
        tools_schema.append(tool_schema)
    return tools_schema



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Union

from loguru import logger


@dataclass
class Chain(ABC):
    """Chain 的基类，定义了 Chain 的基本结构。
    Chain 的主要作用是规范化输入输出，以及规范化输入的格式
    另外就是对调用 LLM 的参数做预处理，例如对输入的文本进行 Prompt 的拼接和补充，对过长的文本进行切割等
    """

    @property
    @abstractmethod
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""

    def _validate_inputs(self, inputs: Dict[str, Any]) -> None:
        """Check that all inputs are present."""
        missing_keys = set(self.input_keys).difference(inputs)
        if missing_keys:
            raise ValueError(f"Missing some input keys: {missing_keys}")

    def _prep_inputs(self, inputs: Union[Dict[str, Any], Any]) -> Dict[str, str]:
        """Validate and prep inputs."""
        if inputs is None or not inputs:
            raise ValueError("Inputs cannot be empty or None.")

        if not isinstance(inputs, dict):
            _input_keys = set(self.input_keys)
            if len(_input_keys) != 1:
                raise ValueError(
                    f"A single string input was passed in, but this chain expects "
                    f"multiple inputs ({_input_keys}). When a chain expects "
                    f"multiple inputs, please call it by passing in a dictionary, "
                    "eg `chain({'foo': 1, 'bar': 2})`"
                )
            inputs = {list(_input_keys)[0]: inputs}
        self._validate_inputs(inputs)
        return inputs

    @abstractmethod
    def _call(self, inputs: Dict[str, Any]) -> str:
        """Execute the chain."""

    def _check_args_kwargs(self, args: Any, kwargs: Any) -> Dict[str, Any]:
        """Check the arguments and keyword arguments."""
        if args and not kwargs:
            if len(args) != 1 or not isinstance(args[0], Union[str, list]):
                raise ValueError("`run` supports only one positional argument of type str or list.")
            return {self.input_keys[0]: args[0]}

        if kwargs and not args:
            for key in kwargs:
                if key not in self.input_keys:
                    raise ValueError(f"Unexpected keyword argument: {key}")
            return kwargs

        if not kwargs and not args:
            return {}

        raise ValueError(
            f"`run` supported with either positional arguments or keyword arguments"
            f" but not both. Got args: {args} and kwargs: {kwargs}."
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Run the chain as text in, text out or multiple variables, text out."""
        try:
            inputs = self._check_args_kwargs(args, kwargs)
            inputs = self._prep_inputs(inputs)
            return self._call(inputs)
        except Exception as e:
            logger.error(f"Caught an exception: {e}")
            raise



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from dataclasses import dataclass
from http import HTTPStatus
from typing import Any, Callable, Dict, List, Optional, Type

from loguru import logger
from pydantic import BaseModel

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config, retry_decorator


class BadRequestError(Exception):
    pass


class Unauthorized(Exception):
    pass


class RequestLimitError(Exception):
    pass


def not_notry_exception(exception: BaseException):
    if isinstance(exception, BadRequestError):
        return False
    if isinstance(exception, Unauthorized):
        return False
    return True


@dataclass
class DashScope(BaseLanguageModel):
    model: str
    use_max_tokens: bool = True
    MAX_TOKENS: int = 6000
    seed: int = 1234

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools)
        try:
            response = self.completion_with_backoff(**kwargs)
        except Exception as e:
            if "Range of input length" in str(e) or "Prompt is too long." in str(e):
                if self.model == "qwen-max":
                    kwargs["model"] = "qwen-max-longcontext"
                else:
                    raise e
                logger.warning(
                    f"Model {self.model} does not support {self._num_tokens(self.messages, self.model)} tokens. Trying again with {kwargs['model']}."
                )
                response = self.completion_with_backoff(**kwargs)
            else:
                raise e

        logger.trace(f"kwargs:{kwargs}\nresponse:{response}")
        return response.output.choices[0].message.content.strip()

    @retry_decorator(not_notry_exception)
    def completion_with_backoff(self, **kwargs):
        response = self.client.Generation.call(**kwargs)

        status_code, code, message = (
            response.status_code,
            response.code,
            response.message,
        )
        if status_code == HTTPStatus.OK:
            return response
        elif status_code == 400:
            # 400 Bad Request
            raise BadRequestError(f"code: {code}, message:{message}")
        elif status_code == 401:
            # 401 Unauthorized
            raise Unauthorized()
        elif status_code == 429:
            # 404 Not Found
            raise RequestLimitError()
        else:
            raise Exception(
                f"Failed request_id: {response.request_id}, status_code: {status_code}, code: {code}, message:{message}"
            )

    @property
    def default_params(self) -> Dict[str, Any]:
        kwargs = {
            "model": self.model,
            "seed": self.seed,
            "result_format": "message",
        }
        if self.use_max_tokens:
            kwargs["max_tokens"] = self.max_tokens
        return kwargs

    def _create_client(self):
        try:
            import dashscope
        except ImportError as e:
            raise ImportError("You need to install `pip install dashscope` to use this provider.") from e
        dashscope.api_key = config.dashscope_api_key
        return dashscope

    def _num_tokens(self, messages: list, model: str):
        if model == "qwen-max" or model == "qwen-max-longcontext":
            logger.trace("qwen-max may change over time. Returning num tokens assuming qwen-turbo.")
            return self._num_tokens(messages, model="qwen-turbo")
        try:
            response = self.client.Tokenization.call(model=model, messages=messages)
        except KeyError:
            logger.trace("model not found. Using qwen-turbo encoding.")
            response = self.client.Tokenization.call(model="qwen-turbo", messages=messages)
        if response.status_code == HTTPStatus.OK:
            return response.usage["input_tokens"]
        else:
            raise Exception(
                f"Failed request_id: {response.request_id}, status_code: {response.status_code}, code: {response.code}, message:{response.message}"
            )

    @property
    def max_tokens(self):
        # add 1000 tokens for answers
        tokens = self._num_tokens(messages=self.messages, model=self.model)
        if not self.MAX_TOKENS > tokens:
            raise Exception(
                f"Prompt is too long. This model's maximum context length is {self.MAX_TOKENS} tokens. your messages required {tokens} tokens"
            )
        if self.model == "qwen-turbo":
            return 1500
        return 2000



import json
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Type

from loguru import logger
from pydantic import BaseModel

from uglychain.llm.tools import tools_schema
from uglychain.utils import config

from .openai_api import ChatGPTAPI


@dataclass
class ChatGPT(ChatGPTAPI):
    api_key: str = config.openai_api_key
    base_url: str = config.openai_api_base
    name: str = "OpenAI"
    use_max_tokens: bool = True
    use_native_tools: bool = field(init=False, default=True)

    def generate(
        self,
        prompt: str = "",
        response_model: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Callable]] = None,
    ) -> str:
        kwargs = self.get_kwargs(prompt, response_model, tools)
        if response_model and self.model in [
            "gpt-3.5-turbo-1106",
            "gpt-4-turbo-preview",
            "gpt-4-1106-preview",
            "gpt-4-0125-preview"
        ]:
            kwargs["response_format"] = {"type": "json_object"}
        try:
            response = self.completion_with_backoff(**kwargs)
        except Exception as e:
            if "maximum context length" in str(e) and self.name == "OpenAI":
                if self.model == "gpt-3.5-turbo-1106":
                    kwargs["model"] = "gpt-3.5-turbo-16k"
                elif self.model == "gpt-4":
                    kwargs["model"] = "gpt-4-turbo-preview"
                else:
                    raise e
                logger.warning(
                    f"Model {self.model} does not support {self._num_tokens(self.messages, self.model)+1000} tokens. Trying again with {kwargs['model']}."
                )
                response = self.completion_with_backoff(**kwargs)
            else:
                raise e

        logger.trace(f"kwargs:{kwargs}\nresponse:{response}")
        if self.use_native_tools and tools and response.choices[0].message.tool_calls:
            result = response.choices[0].message.tool_calls[0].function
            return json.dumps({"name": result.name, "args": json.loads(result.arguments)})
        return response.choices[0].message.content.strip()

    def get_kwargs(
        self,
        prompt: str,
        response_model: Optional[Type],
        tools: Optional[List[Callable]],
    ) -> Dict[str, Any]:
        if self.use_native_tools and tools:
            self._generate_validation()
            self._generate_messages(prompt)
            params = self.default_params
            params["tools"] = tools_schema(tools)
            kwargs = {"messages": self.messages, **params}
            return kwargs
        else:
            return super().get_kwargs(prompt, response_model, tools)

    def _num_tokens(self, messages: list, model: str):
        try:
            import tiktoken
        except ImportError as e:
            raise ImportError("You need to install `pip install tiktoken` to use `use_max_tokens` param.") from e
        if model == "gpt-3.5-turbo" or model == "gpt-3.5-turbo-16k" or model == "gpt-3.5-turbo-1106":
            logger.trace("gpt-3.5-turbo may change over time. Returning num tokens assuming gpt-3.5-turbo-0613.")
            return self._num_tokens(messages, model="gpt-3.5-turbo-0613")
        elif model == "gpt-4" or model == "gpt-4-32k" or model == "gpt-4-1106-preview" or model == "gpt-4-turbo-preview":
            logger.trace("gpt-4 may change over time. Returning num tokens assuming gpt-4-0613.")
            return self._num_tokens(messages, model="gpt-4-0613")
        elif model == "gpt-3.5-turbo-0613":
            # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_message = 4
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif model == "gpt-4-0613":
            tokens_per_message = 3
            tokens_per_name = 1
        else:
            raise NotImplementedError(
                f"""num_tokens() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            logger.trace("model not found. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    @property
    def max_tokens(self):
        tokens = self._num_tokens(messages=self.messages, model=self.model) + 1000  # add 1000 tokens for answers
        if not self.MAX_TOKENS > tokens:
            raise Exception(
                f"Prompt is too long. This model's maximum context length is {self.MAX_TOKENS} tokens. your messages required {tokens} tokens"
            )
        max_tokens = self.MAX_TOKENS - tokens + 1000
        if self.model == "gpt-4-1106-preview":
            return 4096 if max_tokens > 4096 else max_tokens
        return max_tokens



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional

DEFAULT_N = 5


class BaseRetriever(ABC):
    default_n: int = DEFAULT_N

    @abstractmethod
    def search(self, query: str, n: int) -> List[str]:
        pass

    def get(self, query: str) -> str:
        try:
            return self.search(query, 1)[0]
        except Exception:
            return ""


@dataclass
class StoresRetriever(BaseRetriever, ABC):
    path: str
    start_init: bool = False

    def __post_init__(self):
        if self.start_init:
            self.init()
        else:
            self._load()

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def add(self, text: str, metadata: Optional[Dict[str, str]] = None):
        pass

    def _load(self):
        self.init()

    def all(self) -> List[str]:
        return []



import json
import re
from functools import wraps
from typing import Type

from pydantic import BaseModel, ValidationError, create_model

PYDANTIC_FORMAT_INSTRUCTIONS = """
-----
The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{schema}
```

Ensure the response can be parsed by Python json.loads"""


class ParseError(Exception):
    pass


class Instructor(BaseModel):
    @classmethod
    def from_response(cls, response: str) -> "Instructor":
        try:
            # Greedy search for 1st json candidate.
            match = re.search(r"(\{.*\})", response.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
            json_str = ""
            if match:
                json_str = match.group()
            return cls.model_validate_json(json_str)

        except (json.JSONDecodeError, ValidationError) as e:
            name = cls.__name__
            msg = f"Failed to parse {name} from completion {response}. Got: {e}"
            raise ParseError(msg) from e

    @classmethod
    def get_format_instructions(cls) -> str:
        schema = cls.model_json_schema()

        # Remove extraneous fields.
        reduced_schema = schema
        if "title" in reduced_schema:
            del reduced_schema["title"]
        if "type" in reduced_schema:
            del reduced_schema["type"]
        # Ensure json in context is well-formed with double quotes.
        schema_str = json.dumps(reduced_schema)

        return PYDANTIC_FORMAT_INSTRUCTIONS.format(schema=schema_str)

    @classmethod
    def from_BaseModel(cls, cls1) -> Type["Instructor"]:
        if not issubclass(cls1, BaseModel):
            raise TypeError("Class must be a subclass of pydantic.BaseModel")

        return wraps(cls1, updated=())(
            create_model(
                cls1.__name__,
                __base__=(cls1, Instructor),  # type: ignore
            )
        )



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from uglychain.llm import BaseLanguageModel
from uglychain.utils import config

from .baichuan import Baichuan
from .chatgpt import ChatGPT
from .copilot import Copilot
from .custom import Custom
from .dashscope import DashScope
from .gemini import Gemini
from .yi import Yi
from .zhipu import ChatGLM

LLM_PROVIDERS = {
    "gpt-3.5-turbo": (ChatGPT, {"model": "gpt-3.5-turbo-1106", "MAX_TOKENS": 4096}),
    "gpt-3.5-turbo-16k": (ChatGPT, {"model": "gpt-3.5-turbo-16k", "MAX_TOKENS": 16384}),
    "gpt-4": (ChatGPT, {"model": "gpt-4", "MAX_TOKENS": 8192}),
    "gpt-4-turbo": (ChatGPT, {"model": "gpt-4-turbo-preview", "MAX_TOKENS": 4096}),
    "copilot-3.5": (Copilot, {"model": "gpt-3.5-turbo"}),
    "copilot-4": (Copilot, {"model": "gpt-4"}),
    "yi": (Yi, {"model": "yi-34b-chat-v08"}),
    "yi-32k": (Yi, {"model": "yi-34b-chat-32k-v01"}),
    "baichuan-turbo": (Baichuan, {"model": "Baichuan2-Turbo"}),
    "baichuan-turbo-192k": (Baichuan, {"model": "Baichuan2-Turbo-192k"}),
    "baichuan-pro": (Baichuan, {"model": "Baichuan2-53B"}),
    "custom": (Custom, {}),
    "gemini": (Gemini, {"model": "gemini-pro"}),
    "qwen": (DashScope, {"model": "qwen-max", "MAX_TOKENS": 6000}),
    "qwen-turbo": (DashScope, {"model": "qwen-turbo", "MAX_TOKENS": 6000}),
    "qwen-plus": (DashScope, {"model": "qwen-plus", "MAX_TOKENS": 30000}),
    "qwen-28k": (DashScope, {"model": "qwen-max-longcontext", "MAX_TOKENS": 28000}),
    "glm-4": (ChatGLM, {"model": "glm-4", "MAX_TOKENS": 128000}),
    "glm-3": (ChatGLM, {"model": "glm-3-turbo", "MAX_TOKENS": 128000}),
}


def get_llm_provider(llm_provider_name: str = config.llm_provider, is_init_delay: bool = False) -> BaseLanguageModel:
    """
    Get the LLM provider.

    Args:
        llm_provider_name: The name of the LLM provider.

    Returns:
        The LLMProvider object.
    """

    if llm_provider_name in LLM_PROVIDERS.keys():
        provider, kwargs = LLM_PROVIDERS[llm_provider_name]
        kwargs["is_init_delay"] = is_init_delay
        return provider(**kwargs)
    else:
        raise NotImplementedError(f"{llm_provider_name} LLM provider not implemented")


__all__ = ["get_llm_provider"]



from dataclasses import dataclass

from uglychain.utils import config

from .openai_api import ChatGPTAPI


@dataclass
class Custom(ChatGPTAPI):
    api_key: str = config.custom_token
    base_url: str = config.custom_url
    model: str = config.custom_model
    name: str = "CUSTOM"
    use_max_tokens: bool = False



from .chains import LLM, MapChain, ReAct, ReActChain, ReduceChain
from .llm import BaseLanguageModel, Model
from .retrievers import BaseRetriever, StoresRetriever

__version__ = "0.0.4"

__all__ = [
    "BaseLanguageModel",
    "Model",
    "LLM",
    "MapChain",
    "ReduceChain",
    "ReActChain",
    "ReAct",
    "BaseRetriever",
    "StoresRetriever",
]



from .base import Chain
from .llm import LLM, GenericResponseType
from .map import MapChain
from .react import ReAct, ReActChain
from .reduce import ReduceChain

__all__ = [
    "Chain",
    "LLM",
    "MapChain",
    "ReduceChain",
    "ReAct",
    "ReActChain",
    "GenericResponseType",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uglychain',
 'uglychain.chains',
 'uglychain.chains.llm',
 'uglychain.llm',
 'uglychain.provider',
 'uglychain.retrievers',
 'uglychain.utils']

package_data = \
{'': ['*']}

install_requires = \
['docstring-parser>=0.15,<0.16',
 'loguru>=0.7.2,<0.8.0',
 'openai>=1.10.0,<2.0.0',
 'pathos>=0.3.1,<0.4.0',
 'pydantic>=2.5.3,<3.0.0',
 'requests>=2.31.0,<3.0.0',
 'tenacity>=8.2.3,<9.0.0']

setup_kwargs = {
    'name': 'uglychain',
    'version': '0.0.4',
    'description': 'UglyChain：更好用的 LLM 应用构建工具',
    'long_description': '[![Release Notes][release-shield]][release-url]\n[![Contributors][contributors-shield]][contributors-url]\n[![Forks][forks-shield]][forks-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url]\n[![MIT License][license-shield]][license-url]\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n  <a href="https://github.com/uglyboy-tl/UglyChain">\n    <!-- <img src="images/logo.png" alt="Logo" width="80" height="80"> -->\n  </a>\n\n  <h3 align="center">UglyChain</h3>\n\n  <p align="center">\n    ⚡ UglyChain：更好用的 LLM 应用构建工具 ⚡\n    <br />\n    <a href="https://uglychain.uglyboy.cn"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/uglyboy-tl/UglyChain/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/uglyboy-tl/UglyChain/issues">Request Feature</a>\n  </p>\n</div>\n\n## 🤔 What is UglyChain?\n现在有很多利用大模型 LLM 进行应用构建的工具，最有名的就是 LangChain。早期的 LangChain 整个框架并不完善，很多并不直观的定义和调用方式，以及将内部功能封装得太彻底，使的难以定制化的更充分的利用大模型的能力来解决问题。所以我就开发的最初的 UglyGPT（UglyChain的原型），试图解决这个问题。\n\n到了今天，GPTs 也已经面世很长时间了，也有了越来越多的 LLM 应用构建工具。但是这些工具都有一个共同的问题：**不够直观**。\n从底层来说，现在的大模型是基于 Chat 进行接口交互的，这对于应用开发而言并不友好，因为应用开发更多的是模板化的结构化内容生成，而不是对话式的交互。所以我们需要一个对应用开发更加友好的接口，这就是 UglyChain 的初衷。\n\n## Features\n\n- 📦 对大模型接口进行封装，提供对工程化更加直观易懂的交互方式，而不是传统的对话式交互。\n  - 可以参数化 Prompt，更加方便地进行批量调用\n  - 可以对 Prompt 进行结构化返回，方便后续处理\n  - 可以对 Prompt 进行角色设置，方便模型进行角色扮演（这个过程无需操控 Message）\n- 🔗 对大模型的高级调用进行封装，提供更加方便的交互方式\n  - 提供了类似于 MapReduce 的功能，可以通过 MapChain 对多个 Prompt 进行并行调用，也可以用 ReduceChain 对多个 Prompt 进行串行调用\n  - 大模型最优质的能力之一就是拥有 ReAct 能力。我们提供了 ReActChain 便捷的实现这种能力。\n- 💾 提供了搜索引擎的封装，可以方便地进行搜索引擎的调用。\n  - 注意我们只封装了搜索过程的调用，而没有提供搜索引擎的搭建。如果要构建基于 RAG 的应用，需要利用其他的工具完成资料库的建立，而我们只提供对资料库搜索功能的封装。\n\n## Getting Started\n\nWith pip:\n\n```bash\npip install uglychain\n```\n\n## Usage\n\n### LLM\n\n> 这是最基础的模型调用类，其他的高级类也都继承和使用了这个类的基本功能。\n\n快速使用：\n\n```python\nfrom uglychain import LLM, Model\n\nllm = LLM()\nprint(llm("你是谁？")) # 与模型对话，返回字符串的回答\n```\n\n调整基础配置选项：\n\n```python\nllm = LLM(model = Model.YI) # 可以选择更多的模型，如 Model.GPT3_TURBO、Model.GPT4 等等\nllm = LLM(system_prompt = "我想让你担任职业顾问。我将为您提供一个在职业生涯中寻求指导的人，您的任务是帮助他们根据自己的技能、兴趣和经验确定最适合的职业。您还应该对可用的各种选项进行研究，解释不同行业的就业市场趋势，并就哪些资格对追求特定领域有益提出建议。") # 可以对模型设置角色，这样模型就会以这个角色的视角来回答问题。设置的内容保存在 System Message 中。\n```\n\n参数化 prompt：\n\n```python\nllm = LLM(prompt_template = "{object}的{position}是谁？")\nprint(llm(object = "《红楼梦》", position = "作者"))\nprint(llm(object = "上海市", position = "市长"))\n```\n\n对于 prompt 中只有一个参数的情况，可以直接传入参数：\n\n```python\nllm = LLM("介绍一下{object}")\nprint(llm("Transformer"))\n```\n\n结构化返回结果：\n\n```python\nclass UserDetail(BaseModel):\n    name: str\n    age: int\n\nllm = LLM(response_model=UserDetail)\nprint(llm("Extract Jason is 25 years old")) # UserDetail(name=\'Jason\', age=25)\n```\n\n### MapChain\n\n> 这是一个可以并行对同类型 Prompt 进行调用的类，可以大大提高调用效率。\n\n快速使用：\n\n```python\nfrom uglychain import MapChain\n\nllm = MapChain()\nprint(llm([\n        {"input": "How old are you?"},\n        {"input": "What is the meaning of life?"},\n        {"input": "What is the hottest day of the year?"},\n    ]))\n```\n\n类似于 LLM，也可以对 MapChain 进行更高阶的使用：\n\n```python\nclass AUTHOR(BaseModel):\n    name: str = Field(..., description="姓名")\n    introduction: str = Field(..., description="简介")\n\nllm = MapChain(prompt_template="{book}的{position}是谁？", response_model=AUTHOR, map_keys=["book",])\ninput = [\n    "《红楼梦》",\n    "《西游记》",\n    "《三国演义》",\n    "《水浒传》",\n]\nprint(llm(book=input, position="作者"))\n```\n\n## Roadmap\n\n- [x] 增加 FunctionCall 的能力\n- [ ] 增加 Memory 的能力\n\n## Contributing\n\nContributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement".\nDon\'t forget to give the project a star! Thanks again!\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n## License\n\nDistributed under the MIT License. See `LICENSE.txt` for more information.\n\n<!-- MARKDOWN LINKS & IMAGES -->\n<!-- https://www.markdownguide.org/basic-syntax/#reference-style-links -->\n[release-shield]:https://img.shields.io/github/release/uglyboy-tl/UglyChain.svg?style=for-the-badge\n[release-url]: https://github.com/uglyboy-tl/UglyChain/releases\n[contributors-shield]: https://img.shields.io/github/contributors/uglyboy-tl/UglyChain.svg?style=for-the-badge\n[contributors-url]: https://github.com/uglyboy-tl/UglyChain/graphs/contributors\n[forks-shield]: https://img.shields.io/github/forks/uglyboy-tl/UglyChain.svg?style=for-the-badge\n[forks-url]: https://github.com/uglyboy-tl/UglyChain/network/members\n[stars-shield]: https://img.shields.io/github/stars/uglyboy-tl/UglyChain.svg?style=for-the-badge\n[stars-url]: https://github.com/uglyboy-tl/UglyChain/stargazers\n[issues-shield]: https://img.shields.io/github/issues/uglyboy-tl/UglyChain.svg?style=for-the-badge\n[issues-url]: https://github.com/uglyboy-tl/UglyChain/issues\n[license-shield]: https://img.shields.io/github/license/uglyboy-tl/UglyChain.svg?style=for-the-badge\n[license-url]: https://github.com/uglyboy-tl/UglyChain/blob/master/LICENSE.txt',
    'author': 'uglyboy',
    'author_email': 'uglyboy@uglyboy.cn',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass

from uglychain.utils import config

from .openai_api import ChatGPTAPI


@dataclass
class Yi(ChatGPTAPI):
    api_key: str = config.yi_api_key
    base_url: str = "https://api.lingyiwanwu.com/v1"
    name: str = "Yi"
    use_max_tokens: bool = False



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from dataclasses import dataclass

import requests
from loguru import logger
from tenacity import RetryError, before_sleep_log, retry, stop_after_attempt, wait_fixed

from uglychain.utils import config

from .base import BaseRetriever

BING_SEARCH_API_URL = "https://api.bing.microsoft.com/v7.0/search"


@dataclass
class BingRetriever(BaseRetriever):
    def __post_init__(self):
        assert hasattr(config, "bing_subscription_key"), "Bing subscription key not found in config"
        assert isinstance(config.bing_subscription_key, str), "Bing subscription key should be a string"
        assert config.bing_subscription_key, "Bing subscription key should not be empty"

    def search(self, query: str, n: int = BaseRetriever.default_n) -> list[str]:
        try:
            response = self._send_request(query, n)
            json = response.json()
            results = json.get("webPages", {}).get("value", [])
            return [self._format(result) for result in results[:n]] if results else []
        except (ValueError, requests.RequestException, RetryError) as e:
            logger.error(f"Error occurred while sending request or parsing response: {e}")
            return []

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_fixed(1),
        before_sleep=before_sleep_log(logger, "WARNING"),  # type: ignore
    )
    def _send_request(self, query: str, n: int):
        response = requests.get(
            BING_SEARCH_API_URL,
            headers={"Ocp-Apim-Subscription-Key": config.bing_subscription_key},
            params={"q": query, "count": n},
            timeout=5,
        )
        response.raise_for_status()
        return response

    @classmethod
    def _format(cls, result: dict) -> str:
        if not all([result.get("name"), result.get("url"), result.get("snippet")]):
            logger.error(f"Missing key in search result: {result}")
            raise KeyError("Missing key in search result")
        return f"[{result.get('name')}]({result.get('url')})\n> {result.get('snippet')}"



#!/usr/bin/env python3
# -*-coding:utf-8-*-

from .arxiv import ArxivRetriever
from .base import BaseRetriever, StoresRetriever
from .bing import BingRetriever
from .llama_index import LlamaIndexGraphRetriever, LlamaIndexRetriever

try:
    from .bm25 import BM25Retriever
except ImportError:
    BM25Retriever = None

RETRIEVERS = {
    "bing": BingRetriever,
    "arxiv": ArxivRetriever,
    "llama_index": LlamaIndexRetriever,
    "llama_index_graph": LlamaIndexGraphRetriever,
}

STORE_RETRIEVERS = {"bm25": BM25Retriever}


def get_retriever(retriever_name: str = "combine") -> BaseRetriever:
    if retriever_name in RETRIEVERS.keys():
        retriever = RETRIEVERS[retriever_name]
        if retriever is None:
            raise NotImplementedError(f"{retriever_name} Retriever not installed")
        else:
            return retriever()
    else:
        raise NotImplementedError(f"{retriever_name} Retriever not implemented")


def get_stores_retriever(retriever_name: str = "bm25", path: str = "", start_int: bool = False) -> StoresRetriever:
    if retriever_name in STORE_RETRIEVERS.keys():
        retriever = STORE_RETRIEVERS[retriever_name]
        if retriever is None:
            raise NotImplementedError(f"{retriever_name} StoresRetriever not installed")
        else:
            return retriever(path, start_int)
    else:
        raise NotImplementedError(f"{retriever_name} StoresRetriever not implemented")


__all__ = ["BaseRetriever", "StoresRetriever", "get_retriever", "get_stores_retriever"]
